/* @(#)drv_mmc.c	1.2 97/07/18 Copyright 1997 J. Schilling */
#ifndef lint
static	char sccsid[] =
	"@(#)drv_mmc.c	1.2 97/07/18 Copyright 1997 J. Schilling";
#endif
/*
 *	CDR device implementation for
 *	SCSI-3/mmc conforming drives
 *	e.g. Yamaha CDR-400, Ricoh MP6200
 *
 *	Currently using close_track_philips() XXX FIXME
 *
 *	Copyright (c) 1997 J. Schilling
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <standard.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/file.h>

#include <btorder.h>
#include <scgio.h>
#include <scsidefs.h>
#include <scsireg.h>

#include "cdrecord.h"
#include "scsitransp.h"

struct	scg_cmd		scmd;
struct	scsi_inquiry	inq;

extern	int	scsibus;
extern	int	target;
extern	int	lun;

extern	int	silent;
extern	int	verbose;
extern	int	lverbose;

LOCAL	int	mmc_load		__PR((void));
LOCAL	int	mmc_unload		__PR((void));
LOCAL	cdr_t	*identify_mmc		__PR((cdr_t *, struct scsi_inquiry *));
LOCAL	int	speed_select_mmc	__PR((int speed, int dummy));
LOCAL	int	next_wr_addr_mmc	__PR((int track, long *ap));
LOCAL	int	open_track_mmc		__PR((int track, int secsize, int sectype, int tracktype, int dbtype));
LOCAL	int	close_track_mmc		__PR((int track));
LOCAL	int	open_session_mmc	__PR((int toctype, int multi));
LOCAL	int	fixate_mmc		__PR((int onp, int dummy, int toctype));


LOCAL int
mmc_load()
{
	return (scsi_load_unload(1));
}

LOCAL int
mmc_unload()
{
	return (scsi_load_unload(0));
}

cdr_t	cdr_mmc = {
	0, 0,
	"generic SCSI-3/mmc CD-R driver",
	identify_mmc,
	drive_attach,
	drive_getdisktype,
	scsi_load,
/*	mmc_load,*/
	scsi_unload,
	recovery_needed,
	recover,
	speed_select_mmc,
	select_secsize,
	next_wr_addr_mmc,
	reserve_track,
	write_xg1,
	open_track_mmc,
/*	close_track_mmc,*/
	close_track_philips,
	open_session_mmc,
	cmd_dummy,
	read_session_offset,
	fixate_mmc,
};

/*
 * SCSI-3/mmc conformant CD drive
 */
cdr_t	cdr_cd = {
	0, CDR_ISREADER,
	"generic SCSI-3/mmc CD driver",
	identify_mmc,
	drive_attach,
	drive_getdisktype,
	scsi_load,
	scsi_unload,
	cmd_dummy, 				/* recovery_needed	*/
	(int(*)__PR((int)))cmd_dummy,		/* recover		*/
	speed_select_mmc,
	select_secsize,
	(int(*)__PR((int, long *)))cmd_dummy,	/* next_wr_addr		*/
	reserve_track,
	scsi_cdr_write,
	open_track_mmc,
	close_track_mmc,
	(int(*)__PR((int, int)))cmd_dummy,
	cmd_dummy,
	read_session_offset,
	fixation,
};

/*
 * Old pre SCSI-3/mmc CD drive
 */
cdr_t	cdr_oldcd = {
	0, CDR_ISREADER,
	"generic SCSI-2 CD driver",
	identify_mmc,
	drive_attach,
	drive_getdisktype,
	scsi_load,
	scsi_unload,
	cmd_dummy, 				/* recovery_needed	*/
	(int(*)__PR((int)))cmd_dummy,		/* recover		*/
	speed_select_mmc,
	select_secsize,
	(int(*)__PR((int, long *)))cmd_dummy,	/* next_wr_addr		*/
	reserve_track,
	scsi_cdr_write,
	open_track_mmc,
	close_track_mmc,
	(int(*)__PR((int, int)))cmd_dummy,
	cmd_dummy,
	read_session_offset_philips,
	fixation,
};

#if defined(_BIT_FIELDS_LTOH)	/* Intel byteorder */

struct cd_mode_page_05 {		/* write parameters */
	u_char	MP_P_CODE;		/* parsave & pagecode */
	u_char	p_len;			/* 0x32 = 50 Bytes */
	u_char	write_type	:4;	/* Session write type (PACKET/TAO...)*/
	u_char	test_write	:1;	/* Do not actually write data	     */
	u_char	res_2		:3;
	u_char	track_mode	:4;	/* Track mode (Q-sub control nibble) */
	u_char	copy		:1;	/* 1st higher gen of copy prot track ~*/
	u_char	fp		:1;	/* Fixed packed (if in packet mode)  */
	u_char	multi_session	:2;	/* Multi session write type	     */
	u_char	dbtype		:4;	/* Data block type		     */
	u_char	res_4		:4;	/* Reserved			     */
	u_char	res_56[2];		/* Reserved			     */
	u_char	host_appl_code	:6;	/* Host application code of disk     */
	u_char	res_7		:2;	/* Reserved			     */
	u_char	session_format;		/* Session format (DA/CDI/XA)	     */
	u_char	res_9;			/* Reserved			     */
	u_char	packet_size[4];		/* # of user datablocks/fixed packet */
	u_char	audio_pause_len[2];	/* # of blocks where index is zero   */
	u_char	media_cat_number[16];	/* Media catalog Number (MCN)	     */
	u_char	ISRC[14];		/* ISRC for this track		     */
	u_char	sub_header[4];
	u_char	vendor_uniq[4];
};

#else				/* Motorola byteorder */

struct cd_mode_page_05 {		/* write parameters */
	u_char	MP_P_CODE;		/* parsave & pagecode */
	u_char	p_len;			/* 0x32 = 50 Bytes */
	u_char	res_2		:3;
	u_char	test_write	:1;	/* Do not actually write data	     */
	u_char	write_type	:4;	/* Session write type (PACKET/TAO...)*/
	u_char	multi_session	:2;	/* Multi session write type	     */
	u_char	fp		:1;	/* Fixed packed (if in packet mode)  */
	u_char	copy		:1;	/* 1st higher gen of copy prot track */
	u_char	track_mode	:4;	/* Track mode (Q-sub control nibble) */
	u_char	res_4		:4;	/* Reserved			     */
	u_char	dbtype		:4;	/* Data block type		     */
	u_char	res_56[2];		/* Reserved			     */
	u_char	res_7		:2;	/* Reserved			     */
	u_char	host_appl_code	:6;	/* Host application code of disk     */
	u_char	session_format;		/* Session format (DA/CDI/XA)	     */
	u_char	res_9;			/* Reserved			     */
	u_char	packet_size[4];		/* # of user datablocks/fixed packet */
	u_char	audio_pause_len[2];	/* # of blocks where index is zero   */
	u_char	media_cat_number[16];	/* Media catalog Number (MCN)	     */
	u_char	ISRC[14];		/* ISRC for this track		     */
	u_char	sub_header[4];
	u_char	vendor_uniq[4];
};

#endif

#if defined(_BIT_FIELDS_LTOH)	/* Intel byteorder */

struct cd_mode_page_2A {		/* CD Cap / mech status */
	u_char	MP_P_CODE;		/* parsave & pagecode */
	u_char	p_len;			/* 0x14 = 20 Bytes */
	u_char	cd_r_read	:1;	/* Reads CD-R  media		     */
	u_char	cd_rw_read	:1;	/* Reads CD-RW media		     */
	u_char	method2		:1;	/* Reads fixed packet method2 media  */
	u_char	res_2		:5;
	u_char	cd_r_write	:1;	/* Supports writing CD-R  media	     */
	u_char	cd_rw_write	:1;	/* Supports writing CD-RW media	     */
	u_char	test_write	:1;	/* Supports emulation write	     */
	u_char	res_3		:5;	/* Reseved			     */
	u_char	audio_play	:1;	/* Supports Audio play operation     */
	u_char	composite	:1;	/* Deliveres composite A/V stream    */
	u_char	digital_port_2	:1;	/* Supports digital output on port 2 */
	u_char	digital_port_1	:1;	/* Supports digital output on port 1 */
	u_char	mode_2_form_1	:1;	/* Reads Mode-2 form 1 media (XA)    */
	u_char	mode_2_form_2	:1;	/* Reads Mode-2 form 2 media	     */
	u_char	multi_session	:1;	/* Reads multi-session media	     */
	u_char	res_4		:1;	/* Reserved			     */
	u_char	cd_da_supported	:1;	/* Reads audio data with READ CD cmd */
	u_char	cd_da_accurate	:1;	/* READ CD data stream is accurate   */
	u_char	rw_supported	:1;	/* Reads R-W sub channel information */
	u_char	rw_deint_corr	:1;	/* Reads de-interleved R-W sub chan  */
	u_char	c2_pointers	:1;	/* Supports C2 error pointers	     */
	u_char	ISRC		:1;	/* Reads ISRC information	     */
	u_char	UPC		:1;	/* Reads media catalog number (UPC)  */
	u_char	read_bar_code	:1;	/* Supports reading bar codes	     */
	u_char	lock		:1;	/* PREVENT/ALLOW may lock media	     */
	u_char	lock_state	:1;	/* Lock state 0=unlocked 1=locked    */
	u_char	prevent_jumper	:1;	/* State of prev/allow jumper 0=pres */
	u_char	eject		:1;	/* Ejects disc/cartr with STOP LoEj  */
	u_char	res_6_4		:1;	/* Reserved			     */
	u_char	loading_type	:3;	/* Loading mechanism type	     */
	u_char	sep_chan_vol	:1;	/* Vol controls each channel separat */
	u_char	sep_chan_mute	:1;	/* Mute controls each channel separat*/
	u_char	disk_present_rep:1;	/* Changer supports disk present rep */
	u_char	sw_slot_sel	:1;	/* Load empty slot in changer	     */
	u_char	res_7		:4;	/* Reserved			     */
	u_char	max_read_speed[2];	/* Max. read speed in KB/s	     */
	u_char	num_vol_levels[2];	/* # of supported volume levels	     */
	u_char	buffer_size[2];		/* Buffer size for the data in KB    */
	u_char	cur_read_speed[2];	/* Current read speed in KB/s	     */
	u_char	res_16;			/* Reserved			     */
	u_char	res_17_0	:1;	/* Reserved			     */
	u_char	BCK		:1;	/* Data valid on falling edge of BCK */
	u_char	RCK		:1;	/* Set: HIGH high LRCK=left channel  */
	u_char	LSBF		:1;	/* Set: LSB first Clear: MSB first   */
	u_char	length		:2;	/* 0=32BCKs 1=16BCKs 2=24BCKs 3=24I2c*/
	u_char	res_17		:2;	/* Reserved			     */
	u_char	max_write_speed[2];	/* Max. write speed supported in KB/s*/
	u_char	cur_write_speed[2];	/* Current write speed in KB/s	     */
};

#else				/* Motorola byteorder */

struct cd_mode_page_2A {		/* CD Cap / mech status */
	u_char	MP_P_CODE;		/* parsave & pagecode */
	u_char	p_len;			/* 0x14 = 20 Bytes */
	u_char	res_2		:5;
	u_char	method2		:1;	/* Reads fixed packet method2 media  */
	u_char	cd_rw_read	:1;	/* Reads CD-RW media		     */
	u_char	cd_r_read	:1;	/* Reads CD-R  media		     */
	u_char	res_3		:5;	/* Reseved			     */
	u_char	test_write	:1;	/* Supports emulation write	     */
	u_char	cd_rw_write	:1;	/* Supports writing CD-RW media	     */
	u_char	cd_r_write	:1;	/* Supports writing CD-R  media	     */
	u_char	res_4		:1;	/* Reserved			     */
	u_char	multi_session	:1;	/* Reads multi-session media	     */
	u_char	mode_2_form_2	:1;	/* Reads Mode-2 form 2 media	     */
	u_char	mode_2_form_1	:1;	/* Reads Mode-2 form 1 media (XA)    */
	u_char	digital_port_1	:1;	/* Supports digital output on port 1 */
	u_char	digital_port_2	:1;	/* Supports digital output on port 2 */
	u_char	composite	:1;	/* Deliveres composite A/V stream    */
	u_char	audio_play	:1;	/* Supports Audio play operation     */
	u_char	read_bar_code	:1;	/* Supports reading bar codes	     */
	u_char	UPC		:1;	/* Reads media catalog number (UPC)  */
	u_char	ISRC		:1;	/* Reads ISRC information	     */
	u_char	c2_pointers	:1;	/* Supports C2 error pointers	     */
	u_char	rw_deint_corr	:1;	/* Reads de-interleved R-W sub chan  */
	u_char	rw_supported	:1;	/* Reads R-W sub channel information */
	u_char	cd_da_accurate	:1;	/* READ CD data stream is accurate   */
	u_char	cd_da_supported	:1;	/* Reads audio data with READ CD cmd */
	u_char	loading_type	:3;	/* Loading mechanism type	     */
	u_char	res_6_4		:1;	/* Reserved			     */
	u_char	eject		:1;	/* Ejects disc/cartr with STOP LoEj  */
	u_char	prevent_jumper	:1;	/* State of prev/allow jumper 0=pres */
	u_char	lock_state	:1;	/* Lock state 0=unlocked 1=locked    */
	u_char	lock		:1;	/* PREVENT/ALLOW may lock media	     */
	u_char	res_7		:4;	/* Reserved			     */
	u_char	sw_slot_sel	:1;	/* Load empty slot in changer	     */
	u_char	disk_present_rep:1;	/* Changer supports disk present rep */
	u_char	sep_chan_mute	:1;	/* Mute controls each channel separat*/
	u_char	sep_chan_vol	:1;	/* Vol controls each channel separat */
	u_char	max_read_speed[2];	/* Max. read speed in KB/s	     */
	u_char	num_vol_levels[2];	/* # of supported volume levels	     */
	u_char	buffer_size[2];		/* Buffer size for the data in KB    */
	u_char	cur_read_speed[2];	/* Current read speed in KB/s	     */
	u_char	res_16;			/* Reserved			     */
	u_char	res_17		:2;	/* Reserved			     */
	u_char	length		:2;	/* 0=32BCKs 1=16BCKs 2=24BCKs 3=24I2c*/
	u_char	LSBF		:1;	/* Set: LSB first Clear: MSB first   */
	u_char	RCK		:1;	/* Set: HIGH high LRCK=left channel  */
	u_char	BCK		:1;	/* Data valid on falling edge of BCK */
	u_char	res_17_0	:1;	/* Reserved			     */
	u_char	max_write_speed[2];	/* Max. write speed supported in KB/s*/
	u_char	cur_write_speed[2];	/* Current write speed in KB/s	     */
};

#endif

#define	LT_CADDY	0
#define	LT_TRAY		1
#define	LT_POP_UP	2
#define	LT_RES3		3
#define	LT_CHANGER_IND	4
#define	LT_CHANGER_CART	5
#define	LT_RES6		6
#define	LT_RES7		7


struct cd__mode_data {
	struct scsi_mode_header	header;
	union cd_pagex	{
		struct cd_mode_page_05	page05;
		struct cd_mode_page_2A	page2A;
	} pagex;
};

LOCAL cdr_t *
identify_mmc(dp, ip)
	cdr_t			*dp;
	struct scsi_inquiry	*ip;
{
	u_char	mode[0x100];
	int	len;
	struct	cd_mode_page_2A *mp;

	if (ip->type != INQ_WORM && ip->type != INQ_ROMD)
		return ((cdr_t *)0);

	fillbytes((caddr_t)mode, sizeof(mode), '\0');

	silent++;
	if (!get_mode_params(0x2A, "CD capabilities",
			mode, (u_char *)0, (u_char *)0, (u_char *)0, &len)) {
		silent--;
		return (&cdr_cd);
	}
		silent--;

	if (lverbose && len)
		printf("Has CD capabilities (len: %d)\n", len);

	if (len == 0)			/* Pre SCSI-3/mmc drive	 	*/
		return (&cdr_oldcd);

	mp = (struct cd_mode_page_2A *)
		(mode + sizeof(struct scsi_mode_header) +
		((struct scsi_mode_header *)mode)->blockdesc_len);

	if (mp->cd_r_write == 0)	/* SCSI-3/mmc CD drive		*/
		return (&cdr_cd);
	return (dp);			/* Generic SCSI-3/mmc CD-R	*/
}

LOCAL int
speed_select_mmc(speed, dummy)
	int	speed;
	int	dummy;
{
	u_char	mode[0x100];
	int	len;
	long	tmp;
	struct	cd_mode_page_05 *mp;

	fillbytes((caddr_t)mode, sizeof(mode), '\0');

	if (!get_mode_params(0x05, "CD write parameter",
			mode, (u_char *)0, (u_char *)0, (u_char *)0, &len))
		return (-1);
	if (len == 0)
		return (-1);

	mp = (struct cd_mode_page_05 *)
		(mode + sizeof(struct scsi_mode_header) +
		((struct scsi_mode_header *)mode)->blockdesc_len);


	mp->test_write = dummy != 0;

	if (!set_mode_params("CD write parameter", mode, len, 0, -1))
		return (-1);

/*XXX max speed aus page 0x2A lesen !! */

/*	if (scsi_set_speed(-1, speed*176) < 0)*/

	/*
	 * 44100 * 2 * 2 =  176400 bytes/s
	 *
	 * The right formula would be:
	 * tmp = (((long)speed) * 1764) / 10;
	 *
	 * But the standard is rounding the wrong way.
	 * Furtunately rounding down is guaranteed.
	 */
	if (scsi_set_speed(-1, speed*177) < 0)
		return (-1);
	return (0);
}

#if defined(_BIT_FIELDS_LTOH)	/* Intel byteorder */

struct track_info {
	u_char	data_len[2];		/* Data len without this info	*/
	u_char	track_number;		/* Track number for this info	*/
	u_char	session_number;		/* Session number for this info	*/
	u_char	res4;			/* Reserved			*/
	u_char	track_mode	:4;	/* Track mode (Q-sub control)	*/
	u_char	copy		:1;	/* This track is a higher copy	*/
	u_char	damage		:1;	/* if 1 & nwa_valid 0: inc track*/
	u_char	res5_67		:2;	/* Reserved			*/
	u_char	data_mode	:4;	/* Data mode of this track	*/
	u_char	fp		:1;	/* This is a fixed packet track	*/
	u_char	packet		:1;	/* This track is in packet mode	*/
	u_char	blank		:1;	/* This is an invisible track	*/
	u_char	rt		:1;	/* This is a reserved track	*/
	u_char	nwa_valid	:1;	/* Next writable addr valid	*/
	u_char	res7_17		:7;	/* Reserved			*/
	u_char	track_start[4];		/* Track start address		*/
	u_char	next_writable_addr[4];	/* Next writable address	*/
	u_char	free_blocks[4];		/* Free usr blocks in this track*/
	u_char	packet_size[4];		/* Packet size if in fixed mode	*/
	u_char	track_size[4];		/* # of user data blocks in trk	*/
};

#else				/* Motorola byteorder */

struct track_info {
	u_char	data_len[2];		/* Data len without this info	*/
	u_char	track_number;		/* Track number for this info	*/
	u_char	session_number;		/* Session number for this info	*/
	u_char	res4;			/* Reserved			*/
	u_char	res5_67		:2;	/* Reserved			*/
	u_char	damage		:1;	/* if 1 & nwa_valid 0: inc track*/
	u_char	copy		:1;	/* This track is a higher copy	*/
	u_char	track_mode	:4;	/* Track mode (Q-sub control)	*/
	u_char	rt		:1;	/* This is a reserved track	*/
	u_char	blank		:1;	/* This is an invisible track	*/
	u_char	packet		:1;	/* This track is in packet mode	*/
	u_char	fp		:1;	/* This is a fixed packet track	*/
	u_char	data_mode	:4;	/* Data mode of this track	*/
	u_char	res7_17		:7;	/* Reserved			*/
	u_char	nwa_valid	:1;	/* Next writable addr valid	*/
	u_char	track_start[4];		/* Track start address		*/
	u_char	next_writable_addr[4];	/* Next writable address	*/
	u_char	free_blocks[4];		/* Free usr blocks in this track*/
	u_char	packet_size[4];		/* Packet size if in fixed mode	*/
	u_char	track_size[4];		/* # of user data blocks in trk	*/
};

#endif

LOCAL int
next_wr_addr_mmc(track, ap)
	int	track;
	long	*ap;
{
	struct	track_info	track_info;
	long	next_addr;

	if (read_track_info((caddr_t)&track_info, 0xFF, sizeof(track_info)) < 0)
/*	if (read_track_info((caddr_t)&track_info, 0x00, sizeof(track_info)) < 0)*/
		return (-1);
	if (lverbose)
		scsiprbytes("track info:", (u_char *)&track_info,
				sizeof(track_info)-scsigetresid());
	next_addr = a_to_u_long(track_info.next_writable_addr);
	if (ap)
		*ap = next_addr;
	return (0);
}

int	st2mode[] = {
	0,		/* 0			*/
	TM_DATA,	/* 1 ST_ROM_MODE1	*/
	TM_DATA,	/* 2 ST_ROM_MODE2	*/
	0,		/* 3			*/
	0,		/* 4 ST_AUDIO_NOPRE	*/
	TM_PREEM,	/* 5 ST_AUDIO_PRE	*/
	0,		/* 6			*/
	0,		/* 7			*/
};

LOCAL int
open_track_mmc(track, secsize, sectype, tracktype, dbtype)
	int	track;
	int	secsize;
	int	sectype;
	int	tracktype;
	int	dbtype;
{
	u_char	mode[0x100];
	int	len;
	struct	cd_mode_page_05 *mp;

	fillbytes((caddr_t)mode, sizeof(mode), '\0');

	if (!get_mode_params(0x05, "CD write parameter",
			mode, (u_char *)0, (u_char *)0, (u_char *)0, &len))
		return (-1);
	if (len == 0)
		return (-1);

	mp = (struct cd_mode_page_05 *)
		(mode + sizeof(struct scsi_mode_header) +
		((struct scsi_mode_header *)mode)->blockdesc_len);


/*	mp->track_mode = ???;*/
	mp->track_mode = st2mode[sectype & ST_MASK];
/*	mp->copy = ???;*/
	mp->dbtype = dbtype;

/*i_to_short(mp->audio_pause_len, 300);*/
/*i_to_short(mp->audio_pause_len, 150);*/
/*i_to_short(mp->audio_pause_len, 0);*/

	if (!set_mode_params("CD write parameter", mode, len, 0, secsize))
		return (-1);

	return (0);
}

LOCAL int
close_track_mmc(track)
	int	track;
{
	return (0);
}

int	toc2sess[] = {
	SES_DA_ROM,	/* CD-DA		 */
	SES_DA_ROM,	/* CD-ROM		 */
	SES_XA,		/* CD-ROM XA mode 1	 */
	SES_XA,		/* CD-ROM XA MODE 2	 */
	SES_CDI,	/* CDI			 */
	SES_DA_ROM,	/* Invalid - use default */
	SES_DA_ROM,	/* Invalid - use default */
};

LOCAL int
open_session_mmc(toctype, multi)
	int	toctype;
	int	multi;
{
	u_char	mode[0x100];
	int	len;
	struct	cd_mode_page_05 *mp;

	fillbytes((caddr_t)mode, sizeof(mode), '\0');

	if (!get_mode_params(0x05, "CD write parameter",
			mode, (u_char *)0, (u_char *)0, (u_char *)0, &len))
		return (-1);
	if (len == 0)
		return (-1);

	mp = (struct cd_mode_page_05 *)
		(mode + sizeof(struct scsi_mode_header) +
		((struct scsi_mode_header *)mode)->blockdesc_len);


	mp->write_type = WT_TAO;	/* Currently only Track at once */

	mp->multi_session = (multi != 0) ? MS_MULTI : MS_NONE;
	mp->session_format = toc2sess[toctype & TOC_MASK];

	if (!set_mode_params("CD write parameter", mode, len, 0, -1))
		return (-1);

	return (0);
}

LOCAL int
fixate_mmc(onp, dummy, toctype)
	int	onp;
	int	dummy;
	int	toctype;
{
	return (scsi_close_session(0));
}
